import re
import sys
from pathlib import Path
from requests_html import HTMLSession
from bs4 import BeautifulSoup
from requests import Session
from requests.exceptions import ReadTimeout, ConnectionError

PREFIX = """Gemini proxies of a few local news sources:"""

SUFFIX = "\r\n".join([
    "",
    "The code is a work in progress.  The WHYY one sometimes times out "
    "because the remote side is genuinely super slow for some reason.  Usually "
    "a refresh will fix it.",
    "",
    "=> https://tildegit.org/epiii2/pgnp.git Source code repository"])


# I think all three of these are actually wordpress under the hood, and they
# might also be exposing a REST api.  Can I ditch all this HTML parsing and
# just GET JSON?!
# https://whyy.org/wp-json/wp/v2/pages
# Hm, maybe not.  WHYY at least doesn't expose the bulk of the content that
# way.  Still, might help with metadata.  For example, if I'm on some random
# URL like https://whyy.org/nice/ I can see that it's actually "page-id-497788"
# and then request https://whyy.org/wp-json/wp/v2/pages/497788
# There are other types too but I can't find any live examples of them yet.

ROOT = "/proxy/"
TIMEOUT = 8

HEADINGS = ["h%d" % num for num in range(1, 7)]

def first_text(elems):
    """Text of the first element, or an empty string if the list is empty."""
    try:
        return elems[0].text
    except IndexError:
        return ""

def slug(txt):
    """Replace all non-alphanumeric characters with dashes."""
    return re.sub("[^A-Za-z0-9]", "-", txt)

class HandlerOld:
    """The older Handler implementation, based on requests-html."""

    def __init__(self, path):
        self.session = HTMLSession()
        self.path_base = re.sub("/.*$", "", path)
        self.path = re.sub("^[^/?]+", "", path)
        self.root_html = ""
        self.root_gmni = ""
        self.status_code = "20"
        self.ctype = "text/gemini"
        self.output = ""

    def get(self):
        try:
            response = self.session.get(self.root_html + self.path, timeout=TIMEOUT)
        except (ReadTimeout, ConnectionError) as error:
            self.output = "Network error accessing:\r\n\r\n"
            self.output += f"=> {self.root_gmni}{self.path}\r\n"
            response = None
        return response

    def parse(self):
        self.output = f"No handler defined for {self.path_base}\r\n"
        self.output += "=> . Go back to the index\r\n"

    def make_proxy_snippet(self):
        links = [
            (self.root_html + self.path, "HTML Original"),
            (self.root_gmni, "Proxy Home Page")]
        output = "\r\n".join([f"=> {lnk[0]} {lnk[1]}" for lnk in links])
        output += "\r\n\r\n"
        return output

    def render(self):
        sys.stdout.buffer.write(f"{self.status_code} {self.ctype}\r\n".encode())
        if self.status_code == "10" or self.ctype.startswith("text/"):
            sys.stdout.buffer.write(self.output.encode())
        else:
            sys.stdout.buffer.write(self.output)

    def href(self, anchor, attrib="href"):
        try:
            link = anchor.attrs.get(attrib, "")
        except AttributeError:
            link = anchor
        if not link or not self.root_html or not self.root_gmni:
            return link
        if link.startswith("/"):
            return self.root_gmni + link
        else:
            return link.replace(self.root_html, self.root_gmni)


class HandlerDefault(HandlerOld):

    def parse(self):
        self.output += PREFIX + "\r\n\r\n"
        for key in HANDLERS:
            if key:
                self.output += f"=> {ROOT}{key}\r\n"
        self.output += SUFFIX + "\r\n"


class HandlerCitizen(HandlerOld):

    def __init__(self, path):
        super().__init__(path)
        self.root_html = "https://thephiladelphiacitizen.org/"
        self.root_gmni = ROOT + "thephiladelphiacitizen.org/"

    def parse(self):

        r = self.get()
        if not r:
            return

        if r.headers["Content-Type"] in ["image/png", "image/jpeg"]:
            self.ctype = "image/png"
            self.output = r.content
            return

        output = ""
        if r.html.find("body.home"):
            output += self.parse_main_title(r.html)
            output += self.make_proxy_snippet()
            output += self.parse_home_page(r.html)
            output += self.parse_menu(r.html)
        elif r.html.find("body.category") or r.html.find("body.author"):
            output += self.parse_main_title(r.html)
            output += self.make_proxy_snippet()
            output += self.parse_category_page(r.html)
            output += self.parse_menu(r.html)
        elif r.html.find("body.single-post"):
            output += self.make_proxy_snippet()
            output += self.parse_article(r.html)
            output += self.parse_menu(r.html)
        else:
            output += "Unknown page type.  View this via HTTPS instead:\r\n"
            output += f"=> {self.root_html}{self.path}\r\n"

        self.output = output

    def parse_main_title(self, html):
        main_title = first_text(html.find("a#logo h1"))
        return f"# {main_title}\r\n\r\n"

    def parse_home_page(self, html):
        output = ""
        article_links = [(self.href(a), first_text(a.find("h1"))) for a in html.find("div.featuredStory a.featuredStory-titlebar")]
        article_links += [(self.href(a), first_text(a.find("h1"))) for a in html.find("div.storyList a")]
        output += "## Latest Articles\r\n\r\n"
        for url, title in article_links:
            output += f"=> {url} {title}\r\n"
        return output

    def parse_category_page(self, html):
        output = ""
        category_title = first_text(html.find("main header h1"))
        article_links = []
        for a in html.find("main div.row a"):
            url = self.href(a)
            for heading in HEADINGS:
                title = first_text(a.find(heading))
                if title:
                    break
            article_links.append((url, title))
        if category_title:
            output += f"## {category_title}\r\n\r\n"
        for url, title in article_links:
            output += f"=> {url} {title}\r\n"
        output += "\r\n"
        return output

    def parse_article(self, html):
        output = ""
        article_title = first_text(html.find("div.article-headline h1"))
        author = first_text(html.find("div.postAuthor"))
        date = first_text(html.find("div.postDate"))
        output += f"# {article_title}\r\n{author}\r\n{date}\r\n\r\n"
        for elem in html.find("section.mainText article > *"):
            if elem.tag == "p":
                if elem.find("span strong") or elem.find("span b"):
                    output += f"## {elem.text}\r\n\r\n"
                elif elem.find("strong"):
                    if elem.text != "RELATED VIDEO CONTENT":
                        output += f"{elem.text}\r\n\r\n"
                else:
                    output += f"{elem.text}\r\n\r\n"
            elif elem.tag in ["ol", "ul"]:
                for item in elem.find("li"):
                    output += f"* {item.text}\r\n"
                output += "\r\n"
            elif elem.tag == "blockquote":
                # These are repeats of quotes embedded in the article.
                pass
            elif elem.tag == "figure":
                try:
                    imgsrc = self.href(elem.find("img")[0], "src")
                    caption = first_text(elem.find("figcaption"))
                except IndexError:
                    pass
                else:
                    if caption:
                        output += f"=> {imgsrc} Figure: {caption}\r\n\r\n"
                    else:
                        output += f"=> {imgsrc} Figure\r\n\r\n"
            elif elem.tag == "div":
                classes = elem.attrs.get("class", ())
                if "inline_link_container" in classes:
                    # This is a related link.  maybe hold until the end and
                    # then list of links?
                    # TODO
                    pass
                elif "social_footer" in classes:
                    # This is a block of social network links
                    pass
        return output

    def _fallback_link(self, elem, prefix):
        """Kludge for div.voiceWrap that is missing links."""
        a_elems = elem.find("a")
        if a_elems:
            return self.href(a_elems[0])
        else:
            return self.href(prefix + slug(elem.text))

    def parse_menu(self, html):
        menus = {}
        for menudiv in html.find("div.menuContent"):
            key = first_text(menudiv.find("strong"))
            # Evidently most of these links are added after page-load by
            # javascript for some reason.  I think I can kludge it to work
            # without having to restort to javascript!
            if menudiv.attrs.get("id", "") == "voiceWrap":
                links = [(self._fallback_link(li, "/author/"), li.text) for li in menudiv.find("li")]
            else:
                links = [(self.href(a), a.text) for a in menudiv.find("a")]
            if not key in menus:
                menus[key] = []
            menus[key] += links
        output = "## Site Menu\r\n\r\n"
        navlinks = [(self.href(a), a.text) for a in html.find("div.navContainer-wrap a.nothing")]
        for url, title in navlinks:
            output += f"=> {url} {title}\r\n"
        output += "\r\n"
        for section, linkset in menus.items():
            if section:
                output += f"### {section}\r\n\r\n"
            for url, title in linkset:
                output += f"=> {url} {title}\r\n"
            output += "\r\n"
        return output


# Though I just found: https://billypenn.com/wp-json/ !
# and now also https://thephiladelphiacitizen.org/wp-json !!
class HandlerBillyPenn(HandlerOld):

    def __init__(self, path):
        # I don't know why my use of the status code 10 seems to do .../?query
        # instead of .../?s=query but this kludgey hack makes it work like I
        # want
        if "?" in path and "=" not in path:
            path = path.replace("?", "?s=")
        super().__init__(path)
        self.root_html = "https://billypenn.com/"
        self.root_gmni = ROOT + "billypenn.com/"

    def parse(self):

        r = self.get()
        if not r:
            return

        if r.headers["Content-Type"] in ["image/png", "image/jpeg"]:
            self.ctype = "image/png"
            self.output = r.content
            return

        output = ""
        if r.html.find("body.home"):
            output += "# Billy Penn\r\n\r\n"
            output += self.make_proxy_snippet()
            output += self.parse_home_page(r.html)
            output += self.parse_menu(r.html)
        elif r.html.find("body.archive"):
            output += "# Billy Penn\r\n\r\n"
            output += self.make_proxy_snippet()
            output += self.parse_category_page(r.html)
            output += self.parse_menu(r.html)
        elif r.html.find("body.search"):
            output += "# Billy Penn\r\n\r\n"
            parts = r.url.split("?", 1)
            if len(parts) > 1:
                _, query = parts[1].split("=", 1)
                if query:
                    output += self.make_proxy_snippet()
                    output += self.parse_search_page(r.html)
                    output += self.parse_menu(r.html)
                else:
                    output = ""
                    self.status_code = "10"
                    self.ctype = "Search Query"
        elif r.html.find("body.single"):
            output += "# Billy Penn\r\n\r\n"
            output += self.make_proxy_snippet()
            output += self.parse_article(r.html)
            output += self.parse_menu(r.html)
        else:
            output += "Unknown page type.  View this via HTTPS instead:\r\n"
            output += f"=> {self.root_html}{self.path}\r\n"

        self.output = output

    def parse_home_page(self, html):
        output = ""
        article_links = [(self.href(a), a.text) for a in html.find("main article div.stream-item__body h1 a")]
        for url, title in article_links:
            output += f"=> {url} {title}\r\n"
        output += self.parse_paginate_link(html)
        return output

    def parse_search_page(self, html):
        output = ""
        title = first_text(html.find("h1.c-main__title"))
        if title:
            output += f"## {title}\r\n\r\n"
        search_results = first_text(html.find("p.search-tools__total-results"))
        output += search_results + "\r\n\r\n"
        article_links = [(self.href(a), a.text) for a in html.find("main article h1 a")]
        for url, title in article_links:
            output += f"=> {url} {title}\r\n"
        output += self.parse_paginate_link(html)
        return output

    def parse_category_page(self, html):
        output = ""
        category_title = first_text(html.find("main header h1"))
        category_desc = first_text(html.find("h2.c-main__description"))
        if category_title:
            output += f"## {category_title}\r\n"
        if category_desc:
            output += f"{category_desc}\r\n"
        output += "\r\n"
        article_links = [(self.href(a), a.text) for a in html.find("main article h1 a")]
        for url, title in article_links:
            output += f"=> {url} {title}\r\n"
        output += self.parse_paginate_link(html)
        return output

    def parse_paginate_link(self, html):
        output = ""
        page_links = html.find("a[data-ga-category=pagination]")
        if len(page_links) == 1:
            url = self.href(page_links[0])
            title = page_links[0].text
            output += f"\r\n=> {url} {title}\r\n"
        return output

    def parse_article(self, html):
        output = ""
        title = first_text(html.find("article header h1"))
        author = ""
        date = first_text(html.find("article header time"))
        for anchor in html.find("article header a[data-ga-label=author]"):
            if anchor.text:
                author = anchor.text
                break
        if title:
            output += f"# {title}\r\n"
        if author:
            output += f"{author}\r\n"
        if date:
            output += f"{date}\r\n"
        output += "\r\n"

        output += self.parse_header_figure(html)

        for elem in html.find("article section.c-main__body > *"):
            if elem.tag == "p":
                output += f"{elem.text}\r\n\r\n"
            elif elem.tag in HEADINGS:
                prefix = HEADINGS.index(elem.tag) + 1
                prefix = min(4, max(2, prefix))
                prefix = prefix * "#"
                output += f"{prefix} {elem.text}\r\n\r\n"
        return output

    def parse_header_figure(self, html):
        output = ""
        header_figures = html.find("article header figure")
        if header_figures:
            for figure in header_figures:
                try:
                    imgsrc = self.href(figure.find("img")[0], "src")
                    caption_h1 = first_text(figure.find("figcaption h1"))
                    caption_cite = first_text(figure.find("figcaption cite"))
                    if caption_h1:
                        caption = f"{caption_h1} ({caption_cite})"
                    elif caption_cite:
                        caption = f"Photo by {caption_cite}"
                    else:
                        caption = ""
                except IndexError:
                    pass
                else:
                    output += f"=> {imgsrc} {caption}\r\n"
            output += "\r\n"
        return output

    def parse_menu(self, html):
        links = [(self.href(a), a.text) for a in html.find("nav.site-nav--primary li a")]
        output = "## Site Menu\r\n\r\n"
        for url, title in links:
            output += f"=> {url} {title}\r\n"
        return output


class HandlerWHYYNewsOld(HandlerOld):

    def __init__(self, path):
        super().__init__(path)
        self.root_html = "https://whyy.org/"
        self.root_gmni = ROOT + "whyy.org/"

    def parse(self):

        r = self.get()
        if not r:
            return

        if r.headers["Content-Type"] in ["image/png", "image/jpeg"]:
            self.ctype = "image/png"
            self.output = r.content
            return

        output = ""

        output += self.parse_main_title(r.html)
        output += self.make_proxy_snippet()
        output += self.parse_secondary_title(r.html)
        actual_articles = r.html.find("article.article")
        if actual_articles:
            output += self.parse_article(r.html)
        elif r.html.find("article.npr_story_post"):
            output += self.parse_article_npr(r.html)
        else:
            output += self.parse_article_jumble(r.html)

        output += self.parse_menu(r.html)
        self.output = output

    def parse_main_title(self, html):
        output = ""
        title = first_text(html.find("header div.site-logo"))
        if title:
            output = f"# {title}\r\n\r\n"
        return output

    def parse_secondary_title(self, html):
        output = ""
        h1 = first_text(html.find("main header h1"))
        if h1:
            output = f"## {h1}\r\n\r\n"
        return output

    def parse_article_jumble(self, html):
        output = ""
        teasers = html.find("article.content-mode--teaser")
        for teaser in teasers:
            anchor = teaser.find("p.h2-hack a")
            if len(anchor) == 1:
                url = self.href(anchor[0])
                title = anchor[0].text
                output += f"=> {url} {title}\r\n"
        return output

    def parse_article(self, html):
        output = ""
        article = html.find("article.article")[0]
        header = article.find("header")[0]
        author = first_text(header.find("div.entry-meta li"))
        date = first_text(header.find("div.entry-meta span.byline-date"))
        imgsrc = self.href(header.find("figure img")[0], "src")
        caption = first_text(header.find("figure figcaption"))

        article_chunks = []
        for div in article.find("div.detail-content div.flexible-content"):
            classes = div.attrs.get("class", ())
            if "flexible-content--related-content" in classes:
                continue
            article_chunks.append(div.text.replace("\n", " "))

        output += f"By {author} - {date}\r\n\r\n"
        output += f"=> {imgsrc} Figure: {caption}\r\n\r\n"
        for chunk in article_chunks:
            output += chunk + "\r\n\r\n"
        return output

    def parse_article_npr(self, html):
        output = ""
        article = html.find("article.npr_story_post")[0]
        header = article.find("header")[0]
        author = first_text(header.find("div.entry-meta li"))
        date = first_text(header.find("div.entry-meta span.byline-date"))
        imgsrc = self.href(header.find("figure img")[0], "src")
        caption = first_text(header.find("figure figcaption"))
        output += f"By {author} - {date}\r\n\r\n"
        output += f"=> {imgsrc} Figure: {caption}\r\n\r\n"
        output += first_text(
            article.find("div.npr-content")).replace("\n", "\r\n\r\n")
        return output

    def parse_menu(self, html):
        output = ""
        anchors = html.find("ul#menu-header-menu-main a")
        if anchors:
            output = "\r\n## Site Menu\r\n\r\n"
            for anchor in anchors:
                url = self.href(anchor)
                title = anchor.text
                output += f"=> {url} {title}\r\n"
        return output


class Handler:
    """Page handler built on requests and BeautifulSoup instead of requests-html."""

    def __init__(self, path):
        self.session = Session()
        self.path_base = re.sub("/.*$", "", path)
        self.path = re.sub("^[^/?]+", "", path)
        self.root_html = ""
        self.root_gmni = ""
        self.status_code = "20"
        self.ctype = "text/gemini"
        self.output = ""

    def get(self):
        try:
            response = self.session.get(self.root_html + self.path, timeout=TIMEOUT)
        except (ReadTimeout, ConnectionError) as error:
            self.output = "Network error accessing:\r\n\r\n"
            self.output += f"=> {self.root_gmni}{self.path}\r\n"
            response = None
        return response

    def parse(self):
        self.output = f"No handler defined for {self.path_base}\r\n"
        self.output += "=> . Go back to the index\r\n"

    def make_proxy_snippet(self):
        links = [
            (self.root_html + self.path, "HTML Original"),
            (self.root_gmni, "Proxy Home Page")]
        output = "\r\n".join([f"=> {lnk[0]} {lnk[1]}" for lnk in links])
        output += "\r\n\r\n"
        return output

    def render(self):
        sys.stdout.buffer.write(f"{self.status_code} {self.ctype}\r\n".encode())
        if self.status_code == "10" or self.ctype.startswith("text/"):
            sys.stdout.buffer.write(self.output.encode())
        else:
            sys.stdout.buffer.write(self.output)

    def href(self, anchor, attrib="href"):
        try:
            link = anchor.attrs.get(attrib, "")
        except AttributeError:
            link = anchor
        if not link or not self.root_html or not self.root_gmni:
            return link
        if link.startswith("/"):
            return self.root_gmni + link
        else:
            return link.replace(self.root_html, self.root_gmni)


class HandlerWHYYNews(Handler):

    def __init__(self, path):
        super().__init__(path)
        self.root_html = "https://whyy.org/"
        self.root_gmni = ROOT + "whyy.org/"

    def parse(self):

        r = self.get()
        if not r:
            return

        if r.headers["Content-Type"] in ["image/png", "image/jpeg"]:
            self.ctype = "image/png"
            self.output = r.content
            return

        output = ""

        html = BeautifulSoup(r.content, features="lxml")
        output += self.parse_main_title(html)
        output += self.make_proxy_snippet()
        output += self.parse_secondary_title(html)
        actual_articles = html.find_all("article", class_="article")
        if actual_articles:
            output += self.parse_article(html)
        elif html.find_all("article", class_="npr_story_post"):
            output += self.parse_article_npr(html)
        else:
            output += self.parse_article_jumble(html)

        output += self.parse_menu(html)
        self.output = output

    def parse_main_title(self, html):
        output = ""
        for header in html.find_all("header"):
            for logo_div in header.find_all("div", class_="site-logo"):
                for title in logo_div.find_all("a"):
                    title = title.get_text()
                    if title:
                        output = f"# {title}\r\n\r\n"
        return output

    def parse_secondary_title(self, html):
        output = ""
        for main in html.find_all("main"):
            for header in main.find_all("header"):
                for h1 in header.find_all("h1"):
                    h1 = h1.get_text()
                    output = f"## {h1}\r\n\r\n"
        return output

    def parse_article_jumble(self, html):
        output = ""
        for teaser in html.find_all("article", class_="content-mode--teaser"):
            for h2_hack in teaser.find_all("p", class_="h2-hack"):
                anchor = h2_hack.find_all("a")
                if len(anchor) == 1:
                    url = self.href(anchor[0])
                    title = anchor[0].get_text().strip()
                    output += f"=> {url} {title}\r\n"
        return output

    def matching_classes(self, elem, classes):
        """True if any given classes are present for elem, False otherwise."""
        try:
            classes_elem = elem["class"]
        except (TypeError, KeyError):
            return False
        return any(reject in classes_elem for reject in classes)

    def parse_article_meta(self, article):
        header = article.find_all("header")[0]
        for entry_meta in header.find_all("div", class_="entry-meta"):
            for author in entry_meta.find_all("li"):
                author = author.get_text()
            for date in entry_meta.find_all("span", class_="byline-date"):
                date = date.get_text()
        for figure in header.find_all("figure"):
            # They're doing some screwy javascript thing instead of just giving
            # a dang img src
            for img in figure.find_all("img", class_="b-lazy"):
                imgsrc = self.href(img, "data-src")
            for figcaption in figure.find_all("figcaption"):
                caption = figcaption.get_text()
        return (author, date, imgsrc, caption)

    def parse_article(self, html):
        output = ""
        article = html.find_all("article", class_="article")[0]
        author, date, imgsrc, caption = self.parse_article_meta(article)
        article_chunks = []
        rejects = ["block--related-content", "block--doubleclick", "side-ad", "promo-block"]
        for detail_content in article.find_all("div", class_="detail-content"):
            for flex_content in detail_content.find_all("div", class_="flexible-content-container"):
                if self.matching_classes(flex_content, rejects):
                    continue
                for child in flex_content.children:
                    if self.matching_classes(child, rejects):
                        continue
                    try:
                        if child.name == "figure":
                            for img in child.find_all("img"):
                                imgsrc = self.href(img, "src")
                            for cap in child.find_all("figcaption"):
                                caption = cap.get_text()
                            text = f"=> {imgsrc} Figure: {caption}"
                        else:
                            text = child.get_text().replace("\n", " ")
                    except AttributeError:
                        continue
                    if child.name in ["h1", "h2", "h3", "h4", "h5", "h6"]:
                        text = "### " + text
                    article_chunks.append(text)

        output += f"By {author} - {date}\r\n\r\n"
        output += f"=> {imgsrc} Figure: {caption}\r\n\r\n"
        for chunk in article_chunks:
            output += chunk + "\r\n\r\n"
        return output

    def parse_article_npr(self, html):
        output = ""
        article = html.find_all("article", class_="npr_story_post")[0]
        author, date, imgsrc, caption = self.parse_article_meta(article)
        output += f"By {author} - {date}\r\n\r\n"
        output += f"=> {imgsrc} Figure: {caption}\r\n\r\n"
        for content in article.find_all("div", class_="npr-content"):
            output += content.get_text().replace("\n", "\r\n\r\n")
        return output

    def parse_menu(self, html):
        output = ""
        menus = html.find_all("ul", id="menu-header-menu-main")
        if not menus:
            return output
        anchors = menus[0].find_all("a")
        if anchors:
            output = "\r\n## Site Menu\r\n\r\n"
            for anchor in anchors:
                url = self.href(anchor)
                title = anchor.get_text()
                output += f"=> {url} {title}\r\n"
        return output


HANDLERS = {
    "": HandlerDefault,
    "thephiladelphiacitizen.org": HandlerCitizen,
    "billypenn.com": HandlerBillyPenn,
    "whyy.org/news": HandlerWHYYNews
    }
