import re
import sys
import unittest
import pgnp
from pathlib import Path
from unittest.mock import (Mock, create_autospec, DEFAULT)
from requests_html import HTML

class SessionMock(pgnp.handlers.Session):
    """Session with a fake get() method.

    Meant to mock requests.Session.
    Subclass this and set PATH to a local HTML file path.
    """

    PATH = Path("/")

    def get(self, url, **kwargs):
        """Fake a GET request from local disk, with mock response object."""
        url_suffix = re.sub("http[s]://[^/]+/*", "", url)
        response = Mock()
        response.headers = {"Content-Type": "text/html"}
        path = self.__class__.PATH
        if path.exists():
            with open(path, "rb") as f_in:
                response.content = f_in.read()
        else:
            response.content = "404"
        return response

class HTMLSessionMock(pgnp.handlers.HTMLSession):
    """HTMLSession with a fake get() method.

    Meant to mock requests_html.HTMLSession.
    Subclass this and set PATH to a local HTML file path.
    """

    PATH = Path("/")

    def get(self, url, **kwargs):
        """Fake a GET request from local disk, with mock response object."""
        url_suffix = re.sub("http[s]://[^/]+/*", "", url)
        response = Mock()
        response.headers = {"Content-Type": "text/html"}
        path = self.__class__.PATH
        if path.exists():
            with open(path, "rb") as f_in:
                response.content = f_in.read()
        else:
            response.content = "404"
        response.html = HTML(html=response.content)
        return response


class TestBase(unittest.TestCase):
    """A helper parent TestCase (no tests defined).

    Instances of TestBase have a path property defined that points to a
    class-specific directory on disk for supporting files.  This way subclasses
    automatically get supplied with a directory associated with their name for
    their own input and output files.

    On setUp, the HTMLSession within the handlers module is replaced with a
    mock version to load content from input.html within the class' directory.
    On tearDown the original is put back in place.
    """

    @property
    def path(self):
        """Path for supporting files for each class."""
        path = self.__class__.__module__.split(".") + [self.__class__.__name__]
        path.insert(1, "data")
        path = Path("/".join(path))
        return path

    def setUp(self):
        self.maxDiff = None
        self.setup_mock()

    def tearDown(self):
        self.tear_down_mock()

    def setup_mock(self):
        class mocker(SessionMock):
            PATH = self.path / "input.html"
        class mocker_html(HTMLSessionMock):
            PATH = self.path / "input.html"
        hnd = sys.modules["pgnp"].handlers
        hnd.HTMLSessionOrig = hnd.HTMLSession
        hnd.HTMLSession = mocker_html
        hnd.SessionOrig = hnd.Session
        hnd.Session = mocker
    
    def tear_down_mock(self):
        hnd = sys.modules["pgnp"].handlers
        hnd.HTMLSession = hnd.HTMLSessionOrig
        hnd.Session = hnd.SessionOrig
